
#include <boost/tokenizer.hpp>
#include <iostream>
#include "../rand.h"
#include <stdlib.h>
#include <math.h>
#include <set>
#include <sstream>
#include "PolynomialValueFunction.h"
#include "Utility.h"

using namespace std;

PolynomialValueFunction::PolynomialValueFunction()
{
  init();
}

PolynomialValueFunction::~PolynomialValueFunction()
{
  clean();
}

void PolynomialValueFunction::clean()
{
  if(m_poly_weight){
    for(int i = 0; i < m_poly_num_weights; i++)
      delete m_poly_weight[i].index;  
    delete m_poly_weight;
  }
  m_poly_weight = 0;
  m_poly_num_weights = 0;
}

void PolynomialValueFunction::init()
{
  m_poly_weight = 0;
  m_poly_num_weights = 0;
  m_poly_degree= 2;
}

double PolynomialValueFunction::getValue(double* obj, int nobj) const
{
  double val = 0;
  
  for(int i = 0; i < m_poly_num_weights; i++){
    double curr = m_poly_weight[i].w;
    for(int j = 0; j < m_poly_weight[i].degree; j++)
      curr *= obj[m_poly_weight[i].index[j]];
    val += curr;
  }

  return val;
}

bool PolynomialValueFunction::generateWeights(int nobj)
{
  if(!m_poly_num_weights)
    setWeightsNum(2*nobj);
  if(m_poly_degree == 1)
    setWeightsNum(nobj);

  set< multiset<int> > indices_set;
  
  if(!m_poly_num_weights)
    return false;
  
  m_poly_weight = new poly_weight[m_poly_num_weights];

  // at least one weight should be in highest degree
  m_poly_weight[0].w = rndreal(0.1,1);
  m_poly_weight[0].degree = m_poly_degree;
  m_poly_weight[0].index = new int[m_poly_degree];
  multiset<int> indices;  
  for(int j = 0; j < m_poly_weight[0].degree; j++){
    m_poly_weight[0].index[j] = rnd(0,nobj-1);    
    indices.insert(m_poly_weight[0].index[j]);
  }
  indices_set.insert(indices);

  for(int i = 1; i < m_poly_num_weights; i++){
    // provides better spread of values?
    //m_poly_weight[i].w = rndreal(0,1)*rndreal(0,1);
    m_poly_weight[i].w = rndreal(0,1);
    bool novel = false;
    while(!novel){
      indices.clear();
      m_poly_weight[i].degree = rnd(1,m_poly_degree);
      m_poly_weight[i].index = new int[m_poly_weight[i].degree];
      for(int j = 0; j < m_poly_weight[i].degree; j++){
	m_poly_weight[i].index[j] = rnd(0,nobj-1);    
	indices.insert(m_poly_weight[i].index[j]);
      }
      if(indices_set.find(indices) == indices_set.end()){
	indices_set.insert(indices);
	novel = true;
      }
    }
  }

  // normalize weights to sum to one
  double weightsum = 0.;
  for(int i = 0; i < m_poly_num_weights; i++)
    weightsum += m_poly_weight[i].w;
  for(int i = 0; i < m_poly_num_weights; i++)
    m_poly_weight[i].w /= weightsum;
  // approximate weights to second digit
  for(int i = 1; i < m_poly_num_weights; i++)
    m_poly_weight[i].w = (double)floor(m_poly_weight[i].w*100+0.5)/100.;
  m_poly_weight[0].w = 1.;
  for(int i = 1; i < m_poly_num_weights; i++)
    m_poly_weight[0].w -= m_poly_weight[i].w;
  

  // DEBUG
 //  cout << "\nnobj=" << nobj << endl;
//   cout << "\ngenerated polynomial:" << endl;
//   for(int i = 0; i < m_poly_num_weights; i++){
//     cout << m_poly_weight[i].w;
//     for(int j = 0; j < m_poly_weight[i].degree; j++)
//       cout << " * x" << m_poly_weight[i].index[j];
//     cout << endl;
//   }
//   cout << "\npolynomial value:" << endl;
//   double* x = new double[nobj];
//   for(int j = 0; j < nobj; j++){
//     x[j] =  rndreal(0,1);
//     cout << "x" << j << "=" << x[j] << " ";
//   }
//   cout << "v(x)=" << getValue(x, nobj) << endl;
//   delete x;
//   exit(0);

  return true;
}

void PolynomialValueFunction::printPolynomial() const
{
  cout << "\ngenerated polynomial:" << endl;
  cout << toString() << endl;
}

string PolynomialValueFunction::toString() const
{
  ostringstream os;
  
  for(int i = 0; i < m_poly_num_weights; i++){
    if(i)
      os << "+";
    os << m_poly_weight[i].w;
    for(int j = 0; j < m_poly_weight[i].degree; j++)
      os << "*x" << m_poly_weight[i].index[j];
  }
  
  return os.str();
}

void PolynomialValueFunction::readPolynomial(const string& buf) 
{
  typedef boost::tokenizer<boost::char_separator<char> > 
    tokenizer;
  boost::char_separator<char> sep1("+"),sep2("*");
  tokenizer tok1(buf, sep1);

  int num_weights=0;
  for (tokenizer::iterator beg=tok1.begin(); beg!=tok1.end();beg++)
    num_weights++;
  setWeightsNum(num_weights);  
  m_poly_weight = new poly_weight[m_poly_num_weights];

  int i = 0;
  for (tokenizer::iterator beg=tok1.begin(); beg!=tok1.end();beg++,i++){
    tokenizer tok2(*beg, sep2);    
    int degree=-1;
    for (tokenizer::iterator beg2=tok2.begin(); beg2!=tok2.end();beg2++)
      degree++;
    m_poly_weight[i].degree = degree;
    m_poly_weight[i].index = new int[m_poly_weight[i].degree];
    int j = 0;
    tokenizer::iterator beg2 = tok2.begin();
    m_poly_weight[i].w = atof(beg2->c_str());
    for(beg2++; beg2 != tok2.end(); beg2++,j++){
      m_poly_weight[i].index[j] = atoi(beg2->substr(1).c_str());      
    }
  }
}

bool PolynomialValueFunction::setReady(int nobj)
{
  if(!m_poly_weight)
    generateWeights(nobj);
  
  printPolynomial();
  return true;
}
 


double PolynomialValueFunction::getValue(const vector<double>& obj) const
{
  double val = 0;
  
  for(int i = 0; i < m_poly_num_weights; i++){
    double curr = m_poly_weight[i].w;
    for(int j = 0; j < m_poly_weight[i].degree; j++)
      curr *= obj[m_poly_weight[i].index[j]];
    val += curr;
  }

  return val;
}

bool PolynomialValueFunction::acceptableWeights(const Matrix<double>& M, 
						const Matrix<double>& Mean, 
						double thres) const
{
  vector<double> meanvec = Mean.toVector();
  
  int norows = M.GetNoRows();

  Matrix<double> dists(norows,1);
  Matrix<double> vals(norows,1);
  vector<int> dist_sort(norows);
  vector<int> vals_sort(norows);
  
  for(int i = 0; i < norows; i++){
    vector<double> currvec = M.GetRow(i).toVector(); 
    vals(i,0) = getValue(currvec);
    dists(i,0) = eudistance(meanvec,currvec);
  }
  dists.Sort(dist_sort);
  vals.Sort(vals_sort);

  double corr = corrcoeff(dist_sort, vals_sort);
  
  return corr > thres;
}


#ifdef StandalonePolynomialValueFunction
#include <fstream>
int main(int argc, char**argv)
{
  if(argc < 3){
    cerr << "Missing required parameters: <paretofront> <threshold>" << endl;
    exit(0);      
  }

  ifstream in(argv[1]);
  Matrix<double> M(in);

  double corrthres = atof(argv[2]);
  
  int nobj = M.GetNoCols();
  double nex = M.GetNoRows();
  
  Matrix<double> Mean = M.SumR() / nex;

  PolynomialValueFunction ValFunc;
  seed=0.001;
  randomize();

  do{
    ValFunc.clean();
    ValFunc.generateWeights(nobj);
  }
  while(!ValFunc.acceptableWeights(M, Mean, corrthres));

  cout << ValFunc.toString() << endl;
}

#endif
