#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <iostream>
#include "Ranker.h"

using namespace std;

Ranker::Ranker(int* start, int argc, char **argv)
{
  init();
 
  parse_command_line(start, argc,argv);
}

Ranker::~Ranker()
{
  if(m_nadir_points)
    delete[] m_nadir_points;
}


void Ranker::init()
{
  m_use_predictor = false;
  m_queries_number = 0;
  m_curr_iteration = 0;
  m_nadir_points = NULL;
}

bool Ranker::useRanker(int curr_generation) const
{
  if(m_use_gold_preference) 
    return true;
  if(curr_generation < m_skip_num_generations)
    return false;

  return !trainRanker(curr_generation);
}

bool Ranker::trainRanker(int curr_generation) const
{
  if(m_use_gold_preference || m_use_predictor)
    return false;
  
  if(curr_generation < m_skip_num_generations)
    return false;

  return ((curr_generation % m_generations4iteration) == 0);
}

double Ranker::userPreference(double* obj, int objsize) 
{
  m_queries_number++;
  return m_user_preference.getValue(obj, objsize);
}

double Ranker::userPreference(double* obj1, double* obj2, int objsize) 
{
  m_queries_number++;
  return m_user_preference.getPreference(obj1, obj2, objsize);
}


void Ranker::normalizeObjectives(population *pop, int popsize, int objsize)
{
  if(!m_nadir_points)
    estimateNadirPoints(pop, popsize, objsize);

  for(int i = 0; i < popsize; i++)
    for(int j = 0; j < objsize; j++)
      pop->ind[i].obj[j] /= m_nadir_points[j];   
}

void Ranker::estimateNadirPoints(population *pop, int popsize, int objsize)
{
  m_nadir_points = new double[objsize];

  for(int j = 0; j < objsize; j++)
    m_nadir_points[j] = 0.;

  for(int i = 0; i < popsize; i++)
    for(int j = 0; j < objsize; j++)
      if(m_nadir_points[j] < pop->ind[i].obj[j])
	m_nadir_points[j] = pop->ind[i].obj[j];

  for(int j = 0; j < objsize; j++)
    if(m_nadir_points[j] <= 0){
      cerr << "ERROR: nadir point should be > 0, found " 
	   << m_nadir_points[j]
	   << " for objective " << j << endl;
      exit(0);
    }
}

void Ranker::preference_ranking(population *mixed_pop, 
				list* front, int rank, 
				population *new_pop, int popsize, 
				int objsize, int* i)
{
  PreferenceMap prefs;

  //setPreferences(mixed_pop, front, objsize);

  do
     {       
       // set rank for current individual
       mixed_pop->ind[front->index].rank = rank;

       // insert current individual into preference map
       prefs.insert(make_pair(mixed_pop->ind[front->index].preference,front->index));

       // move to next individual
       front = front->child;
     }
   while (front != NULL);

   // copy src individual to destination according to preference ranking
   for(PreferenceMap::const_iterator k = prefs.begin(); k!= prefs.end(); k++,(*i)++){
     if(*i >= popsize)
       break;
     copy_ind (&mixed_pop->ind[k->second], &new_pop->ind[*i]);    
   }
} 

void Ranker::setPreferences(population *pop, list* front, int objsize, bool rankerprefs)
{
   do
     {       
       // compute preference for current individual
       pop->ind[front->index].preference = preference(pop->ind[front->index].obj, objsize, rankerprefs);

       // move to next individual
       front = front->child;
     }
   while (front != NULL);
}

void Ranker::setPreferences(population *pop, int start, int popsize, int objsize, bool rankerprefs)
{
  for(int i = start; i < popsize; i++)
    // compute preference for current individual
    pop->ind[i].preference = preference(pop->ind[i].obj, objsize, rankerprefs);
}

void Ranker::setUserPreferences(population *pop, int objsize, int popsize, int* trainsize)
{
  // init preferences
  for(int i = 0; i < popsize; i++)
    // init preference to maximal
    pop->ind[i].preference = MAXFLOAT;
       

  int size = popsize;
  if(m_max_iteration_examples)
    size = min(size, m_max_iteration_examples);

  if(m_min_allowed_diversity)
    size = min(size, firstRedundant(pop, objsize, size, m_min_allowed_diversity));

  // set preferences
  setPreferences(pop, 0, size, objsize, false);


  *trainsize = size;
}

bool Ranker::update(population *pop, int start, int popsize, int objsize)
{
  bool trained = false;
  double results = 0.;

  if(popsize > start){
    results = train(pop, start, popsize, objsize);
    trained = true;  
  }

  m_curr_iteration++;

  if(m_max_iterations && m_curr_iteration == m_max_iterations){
    cout << "max number of iterations reached" << endl;
    m_use_predictor = true;
  }
  
  if(results >= m_results_threshold){
    cout << "trained ranker satisfies requirements: " << results 
	 << " >= " << m_results_threshold << endl;
    cout << "total number of queries:\t" 
	 << this->getQueriesNumber() << endl;
    m_use_predictor = true;
  }
  else
    cout << "trained ranker does not satisfy requirements: " << results 
	 << " < " << m_results_threshold << endl;    

  return trained;  
}



void Ranker::parse_command_line(int* start, int argc, char **argv)
{
  int i;
  
  m_use_gold_preference = false; 
  m_results_threshold = 0.99;
  m_max_iteration_examples = 0;
  m_max_iterations = 0;
  m_generations4iteration = 10;
  m_skip_num_generations = 200;
  m_min_allowed_diversity = 1E-3;

  // parse options
  for(i=*start;i<argc;i++)
    {
      if(!strcmp(argv[i],"--")){
	i++;
	break;
      }
      if(argv[i][0] != '-') break;
      if(++i>=argc)
	exit_with_help();
      switch(argv[i-1][1])
	{
	case 'U':
	  m_user_preference.setValueFunction(atoi(argv[i]));
	  break;
	case 'G':
	  m_use_gold_preference = true; 
	  i--;
	  break;
	case 'T':
	  m_results_threshold = atof(argv[i]); 
	  break;
	case 'W':
	  m_user_preference.setWeights(argv[i]);
	  break;
	case 'P':
	  m_user_preference.setPolynomial(argv[i]);
	  break;
	case 'D':
	  m_user_preference.setPolynomialDegree(atoi(argv[i]));
	  break;
	case 'N':
	  m_user_preference.setPolynomialWeightsNum(atoi(argv[i]));
	  break;
	case 'M':
	  m_max_iteration_examples = atoi(argv[i]); 
	  break;
	case 'I':
	  m_max_iterations = atoi(argv[i]); 
	  break;
	case 'S':
	  m_skip_num_generations = atoi(argv[i]); 
	  break;
	case 'E':
	  m_min_allowed_diversity = atof(argv[i]); 
	  break;
	case 'B':
	  m_generations4iteration = atoi(argv[i]); 
	  break;
	default:
	  fprintf(stderr,"Unknown option: -%c\n", argv[i-1][1]);
	  exit_with_help();
	}	
    }    
  
  *start = i;
}

void Ranker::exit_with_help()
{
  print_help();
  exit(1);
}

void Ranker::print_help()
{
  printf(
	 "\n User-preference options:\n\n"
	 "-W weights : comma separated list of preference weights  (default=random)\n"	       
	 "-U preference : preference function (default 0)\n"
	 "       0 -- LINEAR\n"
	 "       1 -- QUADRATIC\n"
	 "       2 -- L4\n"
	 "       3 -- TCHEBYCHEFF\n"
	 "       4 -- QUADRATIC_TCHEBYCHEFF\n"
	 "       5 -- POLYNOMIAL\n"
	 "-T threshold: threshold for accepting trained ranker (default=0.99)\n"
	 "-P string: string containing polynomial\n"
	 "-G : always use gold preference function (default false)\n"
         "-D degree: degree of polynomial function (default=2)\n"
         "-N num: number of weights in polynomial function (default=twice number of objectives)\n"
	 "-M num: max number of examples for iteration (default=0 no limit)\n"
         "-I num: max number of training iterations (default=0 no limit)\n"
         "-S num: skip first num generations in training (default=200)\n"
         "-E num: min allowed diversity for keeping training example (default=1E-3)\n"
         "-B num: number of generations between two training iterations (default=10)\n");
}

double distance(double* x, double* y, int size) 
{
  double dist=0.;

  for(int i = 0; i < size; i++)
    dist += (x[i]-y[i])*(x[i]-y[i]);

  return sqrt(dist)/size;
} 


int Ranker::firstRedundant(population *pop, int objsize, int popsize, double min_allowed_diversity)
{
  for(int i = 0; i < popsize-1; i++)
    for(int j = i+1; j < popsize; j++)
    if(distance(pop->ind[i].obj, pop->ind[j].obj, objsize) < min_allowed_diversity)
      return j;
  return popsize;
}
