#include <map>
#include <iostream>
#include <stdlib.h>
#include <octave/oct.h>
#include <math.h>
#include "RegressionRanker.h"


#define Malloc(type,n) (type *)malloc((n)*sizeof(type))
#define Realloc(ptr,type,n) (type *)realloc(ptr,(n)*sizeof(type))


RegressionRanker::RegressionRanker(int start, int argc, char **argv, bool use_priority_vector)
  : Ranker(&start, argc, argv), m_use_priority_vector(use_priority_vector)
{
  init();
  
  const char *error_msg;

  parse_command_line(start, argc, argv);
  error_msg = svm_check_parameter(&m_prob,&m_param);

  if(error_msg)
    {
      fprintf(stderr,"Error: %s\n",error_msg);
      exit(1);
    }
}

RegressionRanker::~RegressionRanker()
{
  if(m_prob.y)
    free(m_prob.y);
  if(m_prob.x)
    free(m_prob.x);
  if(m_x_space)
    free(m_x_space);
  if(m_model)
    svm_destroy_model(m_model);

  svm_destroy_param(&m_param);
}

void RegressionRanker::init()
{
  Ranker::init();
  m_prob.l = 0;
  m_prob.x = NULL;
  m_prob.y = NULL;
  m_model = NULL;
  m_x_space = NULL;
}

double RegressionRanker::preference(double* obj, int objsize, bool rankerprefs) 
{
  double pref;
  
  if(!m_use_gold_preference && rankerprefs){
    svm_node* x = createSvmNode(obj, objsize);
    svm_predict_values(m_model, x, &pref);
    deleteSvmNode(x);
  }
  else{    
    pref = userPreference(obj, objsize);
  }
  return pref;
}


svm_node* RegressionRanker::createSvmNode(double* obj, int objsize)
{
  svm_node* x = Malloc(svm_node,objsize+1);

  for(int i = 0; i < objsize; i++){
    x[i].index = i+1;
    x[i].value = obj[i];
  }
  x[objsize].index = -1;

  return x;
  
}

void RegressionRanker::deleteSvmNode(svm_node* x)
{
  free(x);
}

void RegressionRanker::setPreferences(population *pop, int start, int popsize, int objsize, bool rankerprefs) 
{
  if(!rankerprefs && m_use_priority_vector)
    setPriorityVectorPreferences(pop, start, popsize, objsize);
  else
    Ranker::setPreferences(pop, start, popsize, objsize, rankerprefs);
}



void RegressionRanker::setPriorityVectorPreferences(population *pop, int start, int popsize, int objsize)
{
  // init ranking preferences to 0
  for(int i = start; i < popsize; i++)
    // init preference for current individual
      pop->ind[i].preference = 0;    
      

  for(int i = start; i < popsize; i++)
    for(int j = i+1; j < popsize; j++)
      {
	double pref =  userPreference(pop->ind[i].obj,pop->ind[j].obj, objsize);

	if(pref > 0)
	  pop->ind[i].preference++;
	if(pref < 0)
	  pop->ind[j].preference++;
      }	

  preference2priority(pop, start, popsize);
}

void RegressionRanker::preference2priority(population *pop, int start, int popsize) const
{
  int size = popsize-start;
    
  if(!size)
    return;

  // create comparison matrix
  Matrix A(size,size,1.0);

  for(int r = 0; r < popsize-start; r++){
    for(int c = r+1; c < popsize-start; c++)      
      {
	double diff = pop->ind[r+start].preference - pop->ind[c+start].preference;
	double diffpow = pow(2,fabs(diff));

	if(diff > 0){
	  A(r,c) = diffpow;
	  A(c,r) = 1./diffpow;
	}	  
	else if (diff < 0){
	  A(r,c) = 1./diffpow;
	  A(c,r) = diffpow;	
	}
      }	
  }

  // compute decomposition
  EIG eig = EIG(A);
  
  ComplexColumnVector L = (eig.eigenvalues());
  ComplexMatrix V = (eig.eigenvectors());

  // recover largest eigenvalue
  int pos = 0;
  double value = real(L(0));
  for(int i  = 1; i < size; i++)
    if(real(L(i)) > value){
      value=real(L(i));
      pos = i;
    }
  
  // compute eigevector sum
  double sum = 0;
  for(int i  = 0; i < size; i++)
    sum += real(V(i,pos));

  // assign normalized eigenvector as priority

  for(int i = start; i < popsize; i++)
    pop->ind[i].preference = real(V(i-start,pos))/ sum;  
}


double RegressionRanker::train(population *pop, int start, int popsize, int objsize)
{
  double results = m_results_threshold;

  // add current individuals to training set
  updateSvmProblem(pop, start, popsize, objsize);
	
  std::cout << "\n\ntraining ranker on " << m_prob.l << " examples" << std::endl;

  // clear current model
  if(m_model)
    svm_destroy_model(m_model);

  double p = m_param.p;  
  m_param.p = average_output(m_prob)*p;
  
  double *target = Malloc(double,m_prob.l);
  if(m_do_model_selection){
    //double c = m_param.C;
    //m_param.C = 1;
    results = do_model_selection(m_prob, m_param, target);
    std::cout << "choosing: kernel=" << m_param.kernel_type
	      << " degree=" << m_param.degree
	      << " gamma=" << m_param.gamma << std::endl;
    //m_do_model_selection = false;
    //m_param.C = c;
  }
  else
    // evaluate current parameter setting
    results = do_cross_validation(m_prob, m_param, m_cv_k, target);
  free(target);

  // train model
  m_model = svm_train(&m_prob,&m_param);

  m_param.p = p;

  return results;
}

double RegressionRanker::average_output(const svm_problem& prob) const
{
  double sum = 0.;
  for(int i=0;i<prob.l;i++)
    sum += prob.y[i];
  
  return sum / prob.l;
}

double RegressionRanker::do_model_selection(const svm_problem& prob, svm_parameter& param, double* target)
{
  double best_performance = -MAXFLOAT;
  int folds = (m_cv_k > 0) ? m_cv_k : prob.l; 
  int best_kernel_type = param.kernel_type;
  double best_gamma = param.gamma;
  int best_degree = param.degree;

  //int maxdegree = 4;
  int maxdegree = 2;
  int numgammas = 7;
  double gammas[] = {1e-3,1e-2,1e-1,1,1e1,1e2,1e3};
  
  // linear
  param.kernel_type = 0;
  do_model_evaluation(prob,param, folds, target, &best_performance, &best_kernel_type, &best_degree, &best_gamma);
  if(best_performance >= m_results_threshold)
    return best_performance;

  // polynomial
  param.kernel_type = 1;
  for(int d = 2; d <= maxdegree; d++){
    param.degree = d;
    do_model_evaluation(prob,param, folds, target, &best_performance, &best_kernel_type, &best_degree, &best_gamma);
    if(best_performance >= m_results_threshold)
      return best_performance;
  }

  // gaussian
  param.kernel_type = 2;
  for(int g = 0; g < numgammas; g++){
    param.gamma = gammas[g];
    do_model_evaluation(prob,param, folds, target, &best_performance, &best_kernel_type, &best_degree, &best_gamma);
    if(best_performance >= m_results_threshold)
      return best_performance;
  }
         
  param.kernel_type = best_kernel_type;
  param.degree = best_degree;
  param.gamma = best_gamma;  

  return best_performance;
}

void RegressionRanker::do_model_evaluation(const svm_problem& prob, 
					   const svm_parameter& param, 
					   int k, double* target,
					   double* best_performance,
					   int* best_kernel_type,
					   int* best_degree,
					   double* best_gamma) const
{
  
  double curr_performance = do_cross_validation(prob,param, k, target);	
     
  std::cout << "kernel=" << param.kernel_type
       << " degree=" << param.degree
       << " gamma=" << param.gamma 
       << " cv_performance=" << curr_performance << std::endl;
  
  if(curr_performance > *best_performance){
    *best_performance = curr_performance;
    *best_kernel_type = param.kernel_type;
    *best_degree = param.degree;
    *best_gamma = param.gamma;
  }
}


double RegressionRanker::do_cross_validation(const svm_problem& prob, const svm_parameter& param, int k, double* target) const
{
  int i;
  double total_correct = 0;
  double sumv = 0, sumy = 0, sumvv = 0, sumyy = 0, sumvy = 0;

  srand(1); // always use same splitting
  
  svm_cross_validation(&prob,&param,k,target);

  if(param.svm_type == EPSILON_SVR ||
     param.svm_type == NU_SVR)
    {
      for(i=0;i<prob.l;i++)
	{
	  double y = prob.y[i];
	  double v = target[i];
	  total_correct -= (v-y)*(v-y);
	  sumv += v;
	  sumy += y;
	  sumvv += v*v;
	  sumyy += y*y;
	  sumvy += v*y;
	}

      return ((prob.l*sumvy-sumv*sumy)*(prob.l*sumvy-sumv*sumy))/
	((prob.l*sumvv-sumv*sumv)*(prob.l*sumyy-sumy*sumy));
    }
  else
    {
      for(i=0;i<prob.l;i++)
	if(target[i] == prob.y[i])
	  ++total_correct;    
      return total_correct / prob.l;
    }
}

void RegressionRanker::updateSvmProblem(population *pop, int currstart, int popsize, int objsize)
{
  int start = m_prob.l;

  // update set size
  m_prob.l += popsize-currstart;  

  // allocate memory
  if(m_x_space){
    m_prob.y = Realloc(m_prob.y,double,m_prob.l);
    m_prob.x = Realloc(m_prob.x,struct svm_node *,m_prob.l);
    m_x_space = Realloc(m_x_space,struct svm_node,m_prob.l * (objsize+1));
    // m_prob.x contains obsolete pointers to old x_space!!
    for(int i = 0; i < start; i++)
      m_prob.x[i] = &m_x_space[i*(objsize+1)];
  }
  else{
    m_prob.y = Malloc(double,m_prob.l);
    m_prob.x = Malloc(struct svm_node *,m_prob.l);
    m_x_space = Malloc(struct svm_node,m_prob.l * (objsize+1));
  }

  // fill node and label information
  //std::cout << "\nBEGIN TRAINING SET" << std::endl;
  int j = start*(objsize+1);
  for(int i = currstart; i < popsize; i++){
    m_prob.x[i+start] = &m_x_space[j];
    m_prob.y[i+start] = pop->ind[i].preference;
    //std::cout << m_prob.y[i+start];
    for(int k = 0; k < objsize; k++,j++){
      m_x_space[j].index = k+1;      
      m_x_space[j].value = pop->ind[i].obj[k];
      //std::cout << " " << k+1 << ":" << m_x_space[j].value;
    }
    //std::cout << std::endl;
    m_x_space[j++].index = -1;
  }
  //std::cout << "END TRAINING SET" << std::endl;
}

void print_null(const char *s) {}

void RegressionRanker::parse_command_line(int start, int argc, char **argv)
{
	int i;

	// default values
	m_param.svm_type = EPSILON_SVR;
	m_param.kernel_type = K_LINEAR;
	m_param.degree = 2; // quadratic poly by default
	m_param.gamma = 1;	
	m_param.coef0 = 1; // inhomogeneous poly by default
	m_param.nu = 0.5;
	m_param.cache_size = 100;
	m_param.C = 100;
	m_param.eps = 1e-3;
	m_param.p = 0.1;
	m_param.shrinking = 1;
	m_param.probability = 0;
	m_param.nr_weight = 0;
	m_param.weight_label = NULL;
	m_param.weight = NULL;

	m_do_model_selection = true;
	m_cv_k = 3;

	// parse options
	for(i=start;i<argc;i++)
	{
		if(argv[i][0] != '-') break;
		if(++i>=argc)
			exit_with_help();
		switch(argv[i-1][1])
		{
			case 's':
				m_param.svm_type = atoi(argv[i]);
				break;
			case 't':
				m_param.kernel_type = atoi(argv[i]);
				break;
			case 'd':
				m_param.degree = atoi(argv[i]);
				break;
			case 'g':
				m_param.gamma = atof(argv[i]);
				break;
			case 'r':
				m_param.coef0 = atof(argv[i]);
				break;
			case 'n':
				m_param.nu = atof(argv[i]);
				break;
			case 'm':
				m_param.cache_size = atof(argv[i]);
				break;
			case 'c':
				m_param.C = atof(argv[i]);
				break;
			case 'e':
				m_param.eps = atof(argv[i]);
				break;
			case 'p':
				m_param.p = atof(argv[i]);
				break;
			case 'h':
				m_param.shrinking = atoi(argv[i]);
				break;
			case 'b':
				m_param.probability = atoi(argv[i]);
				break;
			case 'q':
				svm_print_string = &print_null;
				i--;
				break;
			case 'v':
            			m_cv_k = atoi(argv[i]);
				break;
			case 'M':
			        m_do_model_selection = true;
                                i--;
				break;
			case 'w':
				++m_param.nr_weight;
				m_param.weight_label = (int *)realloc(m_param.weight_label,sizeof(int)*m_param.nr_weight);
				m_param.weight = (double *)realloc(m_param.weight,sizeof(double)*m_param.nr_weight);
				m_param.weight_label[m_param.nr_weight-1] = atoi(&argv[i-1][2]);
				m_param.weight[m_param.nr_weight-1] = atof(argv[i]);
				break;
			default:
				fprintf(stderr,"Unknown option: -%c\n", argv[i-1][1]);
				exit_with_help();
		}
	}

// 	if(m_param.svm_type == EPSILON_SVR ||
// 	   m_param.svm_type == NU_SVR){

// 	  m_param.C = max(m_param.C,-1. / m_results_threshold);
// 	  m_param.p = min(m_param.p,-m_results_threshold);
// 	  printf("Setting parameters: C=%f,p=%f\n",m_param.C,m_param.p);
// 	}
}


void RegressionRanker::print_help()
{
  Ranker::print_help();

  printf("\n preference-model options (modified from libsvm 2.89):\n\n");
  printf(
	 "-t kernel_type : set type of kernel function (default 0)\n"
	 "       0 -- linear: u'*v\n"
	 "       1 -- polynomial: (gamma*u'*v + coef0)^degree\n"
	 "       2 -- radial basis function: exp(-gamma*|u-v|^2)\n"
	 "       3 -- sigmoid: tanh(gamma*u'*v + coef0)\n"
	 "       4 -- precomputed kernel (kernel values in training_set_file)\n"
	 "-d degree : set degree in kernel function (default 3)\n"
	 "-g gamma : set gamma in kernel function (default 1/k)\n"
	 "-r coef0 : set coef0 in kernel function (default 0)\n"
	 "-c cost : set the parameter C of C-SVC, epsilon-SVR, and nu-SVR (default 100)\n"
	 "-n nu : set the parameter nu of nu-SVC, one-class SVM, and nu-SVR (default 0.5)\n"
	 "-p epsilon : set the epsilon in loss function of epsilon-SVR (default 0.1)\n"
	 "-m cachesize : set cache memory size in MB (default 100)\n"
	 "-e epsilon : set tolerance of termination criterion (default 0.001)\n"
	 "-h shrinking : whether to use the shrinking heuristics, 0 or 1 (default 1)\n"
	 "-b probability_estimates : whether to train a SVC or SVR model for probability estimates, 0 or 1 (default 0)\n"
	 "-wi weight : set the parameter C of class i to weight*C, for C-SVC (default 1)\n"
	 "-v n: n-fold cross validation (0=loo, default=3)\n"
	 "-M : do model selection (default=true)\n"
	 "-q : quiet mode (no outputs)\n\n"
	 );
}



